//  Copyright (c) 2012 Dennco Project
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

//
//  Created by tkawata on Sep-30, 2012.
//
#include "dcaddcelldialog.h"

#include "dccontainer.h"
#include "dccreator.h"
#include "dcscene.h"
#include "dcvcpage.h"
#include "command/dccommand.h"
#include "utils/dcqtitemmodel.h"
#include "dcsinglecolumntableview.h"
#include "DNUtils.h"

#include <QLineEdit>
#include <QTableView>
#include <QHeaderView>

  DCAddCellDialog::DCAddCellDialog(DCContainer *container, DCCreator *creator, const QString& path, float pageX, float pageY, QWidget *parent) :
      QDialog(parent), d_container(container), d_creator(creator), d_pageX(pageX), d_pageY(pageY)
{
    setWindowTitle(tr("Add cell"));

    d_textField = new QLineEdit;

    d_table = new DCSingleColumnTableView();
    QStringList headers;
    headers << "page";
    d_tableModel = new DCQtItemModel(headers);
    d_tableModel->setReadOnly(0,true);

    d_comboBox = new QComboBox;
    d_comboBox->addItems(QStringList(d_container->getAvailableCellTypes()));
    d_pluginLibraryNameComboBox = new QComboBox;
    d_pluginLibraryNameComboBox->setEditable(true);

    DCScene *scene = container->getScene();
    const QMap<QString,DCVCPage*> *pages = scene->getPages();
    QMapIterator<QString, DCVCPage*> i(*pages);
    int row = 0;
    int selection = -1;
    while (i.hasNext())
    {
        i.next();
        QString item = i.key();
        d_tableModel->insertString(item);
        if (item == path) selection = row;
        row++;
    }
    d_table->setModel(d_tableModel);

    setLayout(new QVBoxLayout);
    d_table->horizontalHeader()->setResizeMode(QHeaderView::Stretch);

    d_okButton = new QPushButton(tr("OK"));
    d_cancelButton = new QPushButton(tr("Cancel"));
    QHBoxLayout *buttonLayout = new QHBoxLayout;
    buttonLayout->addItem(new QSpacerItem(10,10, QSizePolicy::Expanding));
    buttonLayout->addWidget(d_cancelButton);
    buttonLayout->addWidget(d_okButton);
    d_okButton->setEnabled(false);
    d_cancelButton->setAutoDefault(true);
    d_cancelButton->setDefault(true);

    d_pluginInputBox = new QGroupBox();
    QHBoxLayout *pluginInfoLayout = new QHBoxLayout;
    pluginInfoLayout->addWidget(new QLabel(tr("Plugin")));
    pluginInfoLayout->addWidget(d_pluginLibraryNameComboBox);
    d_pluginInputBox->setLayout(pluginInfoLayout);
    d_pluginInputBox->setVisible(false);

    QGridLayout *glayout = new QGridLayout;
    glayout->addWidget(new QLabel(tr("name")),0,0);
    glayout->addWidget(d_textField,0,1);
    glayout->addWidget(new QLabel(tr("type")),1,0);
    glayout->addWidget(d_comboBox, 1, 1);
    glayout->addWidget(d_pluginInputBox, 2,0,1,2);
    ((QVBoxLayout*)layout())->addLayout(glayout);
    layout()->addWidget(d_table);
    d_statusText = new QLabel;
    d_message = new QLabel;
    d_message->setStyleSheet("color : red;");
    layout()->addWidget(d_statusText);
    layout()->addWidget(d_message);
    ((QVBoxLayout*)layout())->addLayout(buttonLayout);

    connect(d_textField, SIGNAL(textChanged(QString)), this, SLOT(textInputChanged(QString)));
    connect(d_okButton, SIGNAL(clicked()), this, SLOT(okButtonClicked()));
    connect(d_cancelButton, SIGNAL(clicked()), this, SLOT(cancelButtonClicked()));
    connect(d_table,SIGNAL(selectionChangedSignal(QItemSelection,QItemSelection)), this, SLOT(listSelectionChanged(QItemSelection,QItemSelection)));
    connect(d_creator, SIGNAL(commandExecuted(const QUndoCommand*)), this, SLOT(commandExecuted(const QUndoCommand*)));
    connect(d_comboBox, SIGNAL(currentIndexChanged(int)), this, SLOT(selectedTypeChanged(int)));
    connect(d_pluginLibraryNameComboBox, SIGNAL(editTextChanged(QString)), this, SLOT(pluginNameChanged(QString)));

    if (selection >= 0)
    {
        QModelIndex index = d_tableModel->index(selection, 0);
        if (index.isValid())
            d_table->setCurrentIndex(index);
    }
}

DCAddCellDialog::~DCAddCellDialog()
{
    if (d_tableModel)
        d_tableModel->deleteLater();
}

bool DCAddCellDialog::checkInput()
{
    bool valid = true;

    d_message->setText("");
    d_statusText->setText("");
    if (d_selectedPagePath.length() == 0)
    {
        d_message->setText(tr("Select a page to place the cell"));
        valid = false;
    }
    else if (d_textField->text().length() > 0)
    {
        QString name;
        if (d_container->getIsPluginType(d_comboBox->currentText()))
        {
            if (d_pluginLibraryNameComboBox->currentText().length() == 0)
            {
                d_message->setText(tr("Input plugin name"));
                valid = false;
            }
            else
            {
                name = d_container->createPluginCellName(d_textField->text(), d_comboBox->currentText(),d_pluginLibraryNameComboBox->currentText());
            }
        }
        else
        {
            name = d_textField->text();
        }

        if (valid)
        {
            std::string path = getFQNString(d_selectedPagePath.toLocal8Bit().data(), name.toLocal8Bit().data());
            TKCell *cell = d_container->getCell(path);
            if (cell)
            {
                d_message->setText(tr("The cell name exists"));
                valid = false;
            }
            else
            {
                d_statusText->setText(QString::fromStdString(path));
            }
        }
    }
    else
    {
        d_message->setText(tr("Input cell name"));
        valid = false;
    }

    if (valid)
    {
        d_okButton->setEnabled(true);
        d_okButton->setDefault(true);
    }
    else
    {
        d_okButton->setEnabled(false);
        d_cancelButton->setDefault(true);
    }
    return valid;
}

void DCAddCellDialog::textInputChanged(const QString &text)
{
    (void)text;

    checkInput();
}

void DCAddCellDialog::okButtonClicked()
{
    if (checkInput())
    {
        QString name;
        if (d_container->getIsPluginType(d_comboBox->currentText()))
        {
            Q_ASSERT(d_pluginLibraryNameComboBox->currentText().length() > 0);
            name = d_container->createPluginCellName(d_textField->text(), d_comboBox->currentText(),d_pluginLibraryNameComboBox->currentText());
        }
        else
        {
            name = d_textField->text();
        }

        d_creator->doCommandAddCell(this, d_container, d_selectedPagePath, name, d_comboBox->currentText(), d_pageX, d_pageY);
    }
}

void DCAddCellDialog::cancelButtonClicked()
{
    done(false);
}


void DCAddCellDialog::listSelectionChanged(const QItemSelection &selected, const QItemSelection &deselected)
{
    (void)deselected;

    if (selected.count()>0)
    {
        d_selectedPagePath = d_tableModel->data(selected.at(0).indexes().at(0), Qt::DisplayRole).toString();
    }
    else
    {
        d_selectedPagePath = "";
    }
    checkInput();
}

void DCAddCellDialog::selectedTypeChanged(int index)
{
    (void)index;

    if (d_container->getIsPluginType(d_comboBox->currentText()))
    {
        d_pluginInputBox->setVisible(true);
        d_pluginLibraryNameComboBox->clear();
        d_pluginLibraryNameComboBox->addItems(d_container->getAvailablePluginLibraries());
    }
    else
    {
        d_pluginInputBox->setVisible(false);
    }
    checkInput();
}

void DCAddCellDialog::pluginNameChanged(const QString &text)
{
    (void)text;
    checkInput();
}

void DCAddCellDialog::commandExecuted(const QUndoCommand *executedCommand)
{
    const DCCommand *command = dynamic_cast<const DCCommand*>(executedCommand);
    if (command && command->getRequester() == this)
    {
        done(true);
    }
}

